/*
 * This is the driver for the matrix-vector multiply tests
 */

/* These are defaults */
#ifndef TIME_EPS
#define TIME_EPS 2.0e-2
#endif
#ifndef MAX_ADAPT_LOOPS
#define MAX_ADAPT_LOOPS 15
#endif
#ifndef MIN_COMMON_TIME
#define MIN_COMMON_TIME 4
#endif
#ifndef MAX_METHOD_NAME 
#define MAX_METHOD_NAME 256
#endif
#define MAX_FLAGS_STR 256

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

/*D
 driver - Run timing test of sparse matrix-vector multiply

 Command line arguments:

. -n  n    - Size of matrix
. -niter n - Number of iterations; these are used to ensure that the operation
            takes long enough to reliably time
. -matrix num - Matrix to use.  Currently,
.vb
        1       5pt operator on 2-d mesh
        2       7pt operator on 3-d mesh
        3      27pt operator on 3-d mesh
        4       dense matrix
        5       read from file
.ve
. -m num  - Paramter for matrix.  Often the mesh size.  This value must be
            consistent with '-n n'
. -nb num - Block size for each mesh point.
. -setup  - Do setup only.  This can be used to isolate the cost of the
            computation.
. -noadapt - Turn off adaptivity of the timing.  This is useful when running
             with performance counters.
. -nocheck - Turn off checking for correctness of the MV implementation
. -concise filename - Append to the given file information about the run
             in a single line, with tab separated data.  

D*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifdef _WIN32
typedef struct { int tv_sec, tv_usec; } time_t;
struct rusage { time_t ru_utime;
                time_t ru_stime;
                int    ru_nvcsw, ru_nivcsw, ru_maxrss;
    };
#define getrusage(a,b)
#define TIMER_SETUP

#elif defined(HAVE_GETRUSAGE)
/* Unix provides a single function for resource usage */
#include <sys/time.h>
#include <sys/resource.h>
#define TIMER_SETUP

#elif defined(HAVE_TIMES) && defined(HAVE_SYSCONF)
#include <unistd.h>
#include <sys/times.h>
typedef struct { int tv_sec, tv_usec; } bsdtime_t;
struct rusage { bsdtime_t ru_utime;
                bsdtime_t ru_stime;
                int    ru_nvcsw, ru_nivcsw, ru_maxrss;
    };
#define getrusage(a,b) {\
		  struct tms aa; double tt;\
		  times(&aa); tt=aa.tms_utime * CLK_FREQ; \
		  (b)->ru_utime.tv_sec = (int)tt;\
                  (b)->ru_utime.tv_usec = (1.0e6)*(tt-(int)tt);\
                  tt=aa.tms_stime * CLK_FREQ;\
		  (b)->ru_stime.tv_sec = (int)tt;\
                  (b)->ru_stime.tv_usec = (1.0e6)*(tt-(int)tt);\
                  (b)->ru_nvcsw=-1;(b)->ru_nivcsw=-1;(b)->ru_maxrss=-1;\
		  }
static double CLK_FREQ;
#if defined(_SC_CLK_TCK)
#define TIMER_SETUP {\
    CLK_FREQ = 1.0 / (double)sysconf(_SC_CLK_TCK);\
    }
#else
'.. you are missing POSIX definitions.  Check your compiler flags .. \
    you may need to add -D_POSIX_SOURCE to your compiler flags'

#endif

#else
#endif

/* Forward declarations */
#ifdef __STDC__
#define ANSI_ARGS(a) a
#else
#define ANSI_ARGS(a) ()
#endif
void ProcessArgs ANSI_ARGS(( int, char **, int *, int *, int *, int *, 
			     int *, int *, int *, int *, FILE ** ));
void CreateMatrix ANSI_ARGS(( int, int, int, int, int **, int **, double **, 
			      int * ));
void CreateVectors ANSI_ARGS(( int, double **, double ** ));

void TestMV ANSI_ARGS(( int *, int *, int *, double *, int ));

#include "mv.h"

int main( argc, argv )
int argc;
char **argv;
{
    double *x, *y, *a;
    char   *p;
    int    *ia, *ja, *ka;
    int    i, j, n, niter, matrix_choice, m_param, n_block, setup_only, nflops,
	   noadapt, nocheck, cnt, cntbest, n_extra;
    double t1, t, tbest;
    FILE   *fp;
    char   method_name[MAX_METHOD_NAME];
    char   method_flags[MAX_FLAGS_STR];
    FILE   *cfp = 0;
    struct rusage r1, r2, r1best, r2best;

    TIMER_SETUP;
    /* Process command line */
    fp = stdout;
    ProcessArgs( argc, argv, &n, &niter, &matrix_choice, &m_param,
		 &n_block, &setup_only, &noadapt, &nocheck, &cfp );

    /* Build Matrix */
    CreateMatrix( matrix_choice, n, m_param, n_block, &ia, &ja, &a, &nflops );

    /* Create vectors */
    CreateVectors( n, &x, &y );

    /* Do any setup */
    n_extra = mvextra_( ia, ja, a, &n );
    ka = 0;
    if (n_extra) ka = (int *)malloc( n_extra * sizeof(int) );
    mvsetup_( ia, ja, ka, a, &n );

    /* Check that code is correct */
    if (! nocheck) 
	TestMV( ia, ja, ka, a, n );

    /* If getting performance counters for initialization only, exit now */
    if (setup_only) return 0;

    /* Run timing test.  This runs the test several times to get
       consistent results unless told not to */
    cnt     = 0;
    tbest   = 100000.0;
    cntbest = 0;
    do {
	getrusage( RUSAGE_SELF, &r1 );
	for (j=0; j<niter; j++) {
	    mv_( ia, ja, ka, a, x, y, &n );
	}
	getrusage( RUSAGE_SELF, &r2 );
	/* Could cycle through data here to flush caches */
	/* Check for consistent results */
	if (! noadapt) {
	    t = r2.ru_utime.tv_sec - r1.ru_utime.tv_sec + 
		1.0e-6 * (r2.ru_utime.tv_usec - r1.ru_utime.tv_usec);
	    if (t < tbest * (1.0 - TIME_EPS)) {
		r1best	= r1;
		r2best	= r2;
		tbest	= t;
		cntbest	= 1;
	    }
	    else if (t < tbest) {
		r1best	= r1;
		r2best	= r2;
		tbest	= t;
		cntbest++;
	    }
	    else if (t < tbest * (1.0 + TIME_EPS)) {
		cntbest++;
	    }
	}
	cnt++;
    } while (!noadapt && cnt < MAX_ADAPT_LOOPS && cntbest < MIN_COMMON_TIME);

    if (!noadapt) {
	r1 = r1best;
	r2 = r2best;
	if (cntbest < MIN_COMMON_TIME) {
	    fprintf( stderr, 
		     "*** Warning: did not get consistent timing results\n" );
	}
    }

    /* Report results */
    mvname_( method_name, MAX_METHOD_NAME );
    mvflags_( method_flags, MAX_FLAGS_STR );

    /* Fortran may have blank-padded.  If so, trim off blanks */
    p = method_name + MAX_METHOD_NAME - 1;
    if (*p == ' ') {
	while (*p == ' ' && p != method_name) p--;
	p[1] = 0;
    }

    /* Fortran may have blank-padded.  If so, trim off blanks */
    p = method_flags + MAX_FLAGS_STR - 1;
    if (*p == ' ') { 
	while (*p == ' ' && p != method_flags) p--;
	p[1] = 0;
    }

    fprintf( fp, "Matrix %d, size %d, param %d:\n", matrix_choice,
	     n, m_param );
    fprintf( fp, "Run of %d iterations with method %s\n", niter, method_name );
    fprintf( fp, "Compiled with %s\n", method_flags );
    t = r2.ru_utime.tv_sec - r1.ru_utime.tv_sec + 
	1.0e-6 * (r2.ru_utime.tv_usec - r1.ru_utime.tv_usec);
    t1 = t;
    t = t / niter;
    fprintf( fp, "User time                    \t %f\n", t );
    t = r2.ru_stime.tv_sec - r1.ru_stime.tv_sec + 
	1.0e-6 * (r2.ru_stime.tv_usec - r1.ru_stime.tv_usec);
    t1 += t;
    t = t / niter;
    fprintf( fp, "System time                  \t %f\n", t );
    t1 = nflops / (t1 / niter);
    fprintf( fp, "MFlops/sec                   \t %f\n", t1 * 1.0e-6 );
    if (r2.ru_nvcsw >= 0)
	fprintf( fp, "Voluntary context switches   \t %d\n", 
		 r2.ru_nvcsw - r1.ru_nvcsw );
    if (r2.ru_nivcsw >= 0)
	fprintf( fp, "Involuntary context switches \t %d\n", 
		 r2.ru_nivcsw - r1.ru_nivcsw );
    if (r2.ru_maxrss >= 0) 
	fprintf( fp, "Max resident set size (pages)\t %d\n", 
		 r2.ru_maxrss );
    fprintf( fp, "Addresses of data\n" );
    fprintf( fp, "\tia\t%lx\n", (long)(char *)ia );
    fprintf( fp, "\tja\t%lx\n", (long)(char *)ja );
    fprintf( fp, "\tka\t%lx\n", (long)(char *)ka );
    fprintf( fp, "\ta\t%lx\n", (long)(char *)a );
    fprintf( fp, "\tx\t%lx\n", (long)(char *)x );
    fprintf( fp, "\ty\t%lx\n", (long)(char *)y );

    if (cfp) {
      fprintf( cfp, "%d\t%d\t%d\t", matrix_choice, n, m_param );
      t = r2.ru_utime.tv_sec - r1.ru_utime.tv_sec + 
	1.0e-6 * (r2.ru_utime.tv_usec - r1.ru_utime.tv_usec);
      t1 = t;
      t = t / niter;
      /* User time */
      fprintf( cfp, "%f\t", t );
      t = r2.ru_stime.tv_sec - r1.ru_stime.tv_sec + 
	1.0e-6 * (r2.ru_stime.tv_usec - r1.ru_stime.tv_usec);
      t1 += t;
      t = t / niter;
      fprintf( cfp, "%f\t", t );
      t1 = nflops / (t1 / niter);
      fprintf( cfp, "%f\t", t1 * 1.0e-6 );
      fprintf( cfp, "%d\t", r2.ru_nvcsw - r1.ru_nvcsw );
      fprintf( cfp, "%d\t", r2.ru_nivcsw - r1.ru_nivcsw );
      fprintf( cfp, "%d\t", r2.ru_maxrss );
      fprintf( cfp, "%s\t%s\n", method_name, method_flags );
      fclose( cfp ); cfp = 0;
    }
    return 0;
}

void ProcessArgs( argc, argv, n_ptr, niter_ptr, matrix_choice_ptr, 
		  m_param_ptr, n_block_ptr, setup_only_ptr, noadapt_ptr,
		  nocheck_ptr, cfp_ptr )
int argc, *n_ptr, *niter_ptr, *matrix_choice_ptr, *m_param_ptr,
    *n_block_ptr, *setup_only_ptr, *noadapt_ptr, *nocheck_ptr;
FILE **cfp_ptr;
char **argv;
{
    /* Set default values */
    *n_ptr	       = 100;
    *niter_ptr	       = 100;
    *matrix_choice_ptr = 1;
    *m_param_ptr       = 10;
    *setup_only_ptr    = 0;
    *n_block_ptr       = 1;
    *noadapt_ptr       = 0;
    *nocheck_ptr       = 0;
    *cfp_ptr           = 0;

    /* Skip Program Name */
    argc--;
    argv++;

    /* Loop through arguments */
    while (argc > 0) {
	if (strcmp( argv[0], "-n" ) == 0) {
	    argc--;
	    argv++;
	    *n_ptr = atoi( argv[0] );
	}
	else if (strcmp( argv[0], "-niter" ) == 0) {
	    argc--;
	    argv++;
	    *niter_ptr = atoi( argv[0] );
	}
	else if (strcmp( argv[0], "-matrix" ) == 0) {
	    argc--;
	    argv++;
	    *matrix_choice_ptr = atoi( argv[0] );
	}
	else if (strcmp( argv[0], "-m" ) == 0) {
	    argc--;
	    argv++;
	    *m_param_ptr = atoi( argv[0] );
	}
	else if (strcmp( argv[0], "-nb" ) == 0) {
	    argc--;
	    argv++;
	    *n_block_ptr = atoi( argv[0] );
	}
	else if (strcmp( argv[0], "-setup" ) == 0) {
	    *setup_only_ptr = 1;
	}
	else if (strcmp( argv[0], "-noadapt" ) == 0) {
	    *noadapt_ptr = 1;
	}
	else if (strcmp( argv[0], "-nocheck" ) == 0) {
	    *nocheck_ptr = 1;
	}
	else if (strcmp( argv[0], "-concise" ) == 0) {
	    argc--;
	    argv++;
	    /* Open file for appending */
	    *cfp_ptr = fopen( argv[0], "w+" );
	    if (!*cfp_ptr) {
	        fprintf( stderr, "Could not open %s\n", argv[0] );
		exit( 1 );
	    }
	}
	else {
	    fprintf( stderr, "Unrecognized argument %s\n", argv[0] );
	    exit( 1 );
	}
	argc--;
	argv++;
    }
}

void Create5pt( m_param, n, n_block, ia_ptr, ja_ptr, a_ptr, nflops_ptr )
int    m_param, n, n_block;
int    **ia_ptr, **ja_ptr, *nflops_ptr;
double **a_ptr;
{
    int    *ia, *ja;
    double *a, sum;
    int    i, j, k, k1, nflops, max_mem, offset, row, cnt;

    if (m_param * m_param * n_block != n) {
	fprintf( stderr, "Mesh size %d inconsistent with matrix size %d\n",
		 m_param, n );
	exit( 1 );
    }
    max_mem = n * 5 * n_block;
    ia = (int *)malloc( (n+1) * sizeof(int) );
    ja = (int *)malloc( (n*5*n_block) * sizeof(int) );
    a  = (double *)malloc( (n*5*n_block) * sizeof(double) );
    if (!ia || !ja || !a) {
	fprintf( stderr, "Could not allocate matrix of size %d\n", n );
	exit( 1 );
    }

    nflops = 0;
    row    = 0;
    cnt    = 0;
    offset = 1;  /* This is needed for Fortran-oriented operations */
    /* i and j are loops over mesh points. 
       k is the loop over the block */
    for (j=0; j<m_param; j++) {
	for (i=0; i<m_param; i++) {
	    for (k=0; k<n_block; k++) {
		ia[row] = cnt + offset;
		/* Compute row sum FIRST */
		sum = -1.0 * (n_block - 1);
		if (j > 0) sum += -1.0 * n_block;
		if (i > 0) sum += -1.0 * n_block; 
		if (j < m_param - 1) sum += -1.0 * n_block;
		if (i < m_param - 1) sum += -1.0 * n_block;
		/* Add in the 5 point ops */
		if (j > 0) {
		    for (k1 = 0;k1<n_block; k1++) {
			ja[cnt] = row - k + k1 - m_param + offset;
			a[cnt]  = -1.0;
			cnt++;
		    }
		}
		if (i > 0) {
		    for (k1 = 0;k1<n_block; k1++) {
			ja[cnt] = row - k + k1 - 1 + offset;
			a[cnt]  = -1.0;
			cnt++;
		    }
		}
		/* Always add diagonal */
		for (k1 = 0;k1<n_block; k1++) {
		    ja[cnt] = row - k + k1 + offset;
		    /* We make the sum 1, rather than zero */
		    if (k1 == k)             /* diagonal */
			a[cnt] = 1.0 - sum;
		    else 
			a[cnt]  = -1.0;
		    cnt++;
		}
		if (i < m_param - 1) {
		    for (k1 = 0;k1<n_block; k1++) {
			ja[cnt] = row - k + k1 + 1 + offset;
			a[cnt]  = -1.0;
			cnt++;
		    }
		}
		if (j < m_param - 1) {
		    for (k1 = 0;k1<n_block; k1++) {
			ja[cnt] = row - k + k1 + m_param + offset;
			a[cnt]  = -1.0;
			cnt++;
		    }
		}
		if (cnt > max_mem) {
		    fprintf( stderr, "Internal error in CreateMatrix(%d)\n",
			     cnt );
		    exit( 1 );
		}
		row++;
	    }
	}
    }
    ia[row] = cnt + offset;
    /* 
       Each matrix element is used once in the matrix-vector product, 
       the number of multiples cnt; adds is cnt - n (first add on each
       row is unnecessary) 
     */
    nflops = 2 * cnt - n; 

    *ia_ptr	= ia;
    *ja_ptr	= ja;
    *a_ptr	= a;
    *nflops_ptr	= nflops;
}

void Create7pt( m_param, n, n_block, ia_ptr, ja_ptr, a_ptr, nflops_ptr )
int    m_param, n, n_block;
int    **ia_ptr, **ja_ptr, *nflops_ptr;
double **a_ptr;
{
    int    *ia, *ja;
    double *a, sum;
    int    i, j, k, kb, kb1, nflops, max_mem, offset, row, cnt;

    if (m_param * m_param * m_param * n_block != n) {
	fprintf( stderr, "Mesh size %d inconsistent with matrix size %d\n",
		 m_param, n );
	exit( 1 );
    }
    max_mem = n * 7 * n_block;
    ia = (int *)malloc( (n+1) * sizeof(int) );
    ja = (int *)malloc( (n*7*n_block) * sizeof(int) );
    a  = (double *)malloc( (n*7*n_block) * sizeof(double) );
    if (!ia || !ja || !a) {
	fprintf( stderr, "Could not allocate matrix of size %d\n", n );
	exit( 1 );
    }

    nflops = 0;
    row    = 0;
    cnt    = 0;
    offset = 1;  /* This is needed for Fortran-oriented operations */
    /* i, j, and k are loops over mesh points. 
       kb is the loop over the block */
    for (k=0; k<m_param; k++) {
	for (j=0; j<m_param; j++) {
	    for (i=0; i<m_param; i++) {
		for (kb=0; kb<n_block; kb++) {
		    ia[row] = cnt + offset;
		    /* Compute row sum FIRST */
		    sum = -1.0 * (n_block - 1);
		    if (j > 0) sum += -1.0 * n_block;
		    if (i > 0) sum += -1.0 * n_block; 
		    if (k > 0) sum += -1.0 * n_block; 
		    if (j < m_param - 1) sum += -1.0 * n_block;
		    if (i < m_param - 1) sum += -1.0 * n_block;
		    if (k < m_param - 1) sum += -1.0 * n_block;
		    /* Add in the 7 point ops */
		    if (k > 0) {
			for (kb1 = 0;kb1<n_block; kb1++) {
			    ja[cnt] = row - kb + kb1 - m_param * m_param + 
				offset;
			    a[cnt]  = -1.0;
			    cnt++;
			}
		    }
		    if (j > 0) {
			for (kb1 = 0;kb1<n_block; kb1++) {
			    ja[cnt] = row - kb + kb1 - m_param + offset;
			    a[cnt]  = -1.0;
			    cnt++;
			}
		    }
		    if (i > 0) {
			for (kb1 = 0;kb1<n_block; kb1++) {
			    ja[cnt] = row - kb + kb1 - 1 + offset;
			    a[cnt]  = -1.0;
			    cnt++;
			}
		    }
		    /* Always add diagonal */
		    for (kb1 = 0;kb1<n_block; kb1++) {
			ja[cnt] = row - kb + kb1 + offset;
			/* We make the sum 1, rather than zero */
			if (kb1 == kb)             /* diagonal */
			    a[cnt] = 1.0 - sum;
			else 
			    a[cnt]  = -1.0;
			cnt++;
		    }
		    if (i < m_param - 1) {
			for (kb1 = 0;kb1<n_block; kb1++) {
			    ja[cnt] = row - kb + kb1 + 1 + offset;
			    a[cnt]  = -1.0;
			    cnt++;
			}
		    }
		    if (j < m_param - 1) {
			for (kb1 = 0;kb1<n_block; kb1++) {
			    ja[cnt] = row - kb + kb1 + m_param + offset;
			    a[cnt]  = -1.0;
			    cnt++;
			}
		    }
		    if (k < m_param - 1) {
			for (kb1 = 0;kb1<n_block; kb1++) {
			    ja[cnt] = row - kb + kb1 + m_param * m_param + 
				offset;
			    a[cnt]  = -1.0;
			    cnt++;
			}
		    }
		    if (cnt > max_mem) {
			fprintf( stderr, 
				 "Internal error in CreateMatrix(%d)\n",
				 cnt );
			exit( 1 );
		    }
		    row++;
		}
	    }
	}
    }
    ia[row] = cnt + offset;
    /* 
       Each matrix element is used once in the matrix-vector product, 
       the number of multiples cnt; adds is cnt - n (first add on each
       row is unnecessary) 
     */
    nflops = 2 * cnt - n; 

    *ia_ptr	= ia;
    *ja_ptr	= ja;
    *a_ptr	= a;
    *nflops_ptr	= nflops;
}

void Create27pt( m_param, n, n_block, ia_ptr, ja_ptr, a_ptr, nflops_ptr )
int    m_param, n, n_block;
int    **ia_ptr, **ja_ptr, *nflops_ptr;
double **a_ptr;
{
    fprintf( stderr, "Not yet implemented\n" );
    exit( 1 );
}

void CreateMatrix( matrix_choice, n, m_param, n_block, 
		   ia_ptr, ja_ptr, a_ptr, nflops_ptr )
int matrix_choice, m_param, n, n_block;
int **ia_ptr, **ja_ptr, *nflops_ptr;
double **a_ptr;
{
    switch (matrix_choice) {
    case 1:
	Create5pt( m_param, n, n_block, ia_ptr, ja_ptr, a_ptr, nflops_ptr );
	break;
    case 2:
	Create7pt( m_param, n, n_block, ia_ptr, ja_ptr, a_ptr, nflops_ptr );
	break;
    case 3:
	Create27pt( m_param, n, n_block, ia_ptr, ja_ptr, a_ptr, nflops_ptr );
	break;
    default:
	fprintf( stderr, "Unrecognized matrix choice %d\n", matrix_choice );
	exit( 1 );
    }
}

void CreateVectors( n, x_ptr, y_ptr )
int n;
double **x_ptr, **y_ptr;
{
    int i;
    double *x, *y;

    /* Provide for some offsets in the allocation */
    *x_ptr = x = (double *) malloc( (n + 16) * sizeof(double) );
    *y_ptr = y = (double *) malloc( (n + 16) * sizeof(double) );
    
    if (!x || !y) {
	fprintf( stderr, "Could not allocate vectors of size %d\n", n );
	exit( 1 );
    }

    /* Initialize with values that will grow slowly in operation */
    for (i=0; i<n+16; i++) {
	x[i] = 1.000031004;
	y[i] = 1.000013303;
    }
}

void TestMV( ia, ja, ka, a, n )
int    *ia, *ja, *ka;
double *a;
int    n;
{
    double *x, *y;
    int    i, errcnt = 0;

    x = (double *)malloc( n * sizeof(double) );
    y = (double *)malloc( n * sizeof(double) );

    if (!x || !y) {
	fprintf( stderr, "Could not allocate vectors of size %d\n", n );
	exit( 1 );
    }

    /* Initialize with values that will grow slowly in operation */
    for (i=0; i<n; i++) {
	x[i] = 1.0;
	y[i] = 0.0;
    }

    /* Check row sum */
    mv_( ia, ja, ka, a, x, y, &n );

    for (i=0; i<n; i++) {
	if (fabs( y[i] - 1.0 ) > 1.0e-10) {
	    errcnt ++;
	    fprintf( stderr, "y[%d] = %e; should be 1.0\n", i, y[i] );
	    if (errcnt > 10) {
		fprintf( stderr, "Too many errors; aborting!\n" );
		exit( 1 );
	    }
	}
    }

#ifdef FOO
    /* Check row n/2 */
    for (i=0; i<n; i++) 
	x[i] = 0.0;

    x[n/2] = 1.0;

    mv_( ia, ja, a, x, y, &n );

    /* Check that y is the n/2 column of y */
    for (i=0; i<n; i++) {
    }
#endif
    
    free( x );
    free( y );

    if (errcnt > 0) {
	fprintf( stderr, "Fix MV before running tests\n" );
	exit( 1 );
    }
}
